// Onur G. Guleryuz 1995, 1996, 1997,
// University of Illinois at Urbana-Champaign,
// Princeton University,
// Polytechnic University.

#include <stdio.h>
#include <stdlib.h>
#include "wav_filters_extern.h"
#include "wav_gen.h"
#include "alloc.h"

// Extend data by ofs pixels on both ends using periodic 
// symmetry. data should be "pointing in" to allow up to data[-ofs]
// and data[N-1+ofs]. 
// ofs>=N is handled.

void extend_periodic(float *data,int N){
/*	int i,k;
	k=N-1;
	for(i=1;i<=2;i++) {
		data[-i]=data[N-i];
		data[k+i]=data[i-1];
	}*/

	data[-1] = data[N-1];
	data[-2] = data[N-2];
	
	data[N] = data[0];
	data[N+1] = data[1];
}

// This routine is used in evaluating forward wavelet transforms.
//
// data of dimension N contains the input.
// float *filter contains the Nf filter taps.
// ofs is the decimation factor.
//
// filtering starts at beg and continues until N+beg, yielding N/ofs samples. 
// data needs to be padded suitably at both ends.
// This hoopla about the starting point is in place to ensure 
// correct boundary processing when evaluating wavelet transforms.
//
// coef returns the result of the filtering and decimation.
// 
// Viewed as a transform the basis function that generates the
// coefficient of index 0 is "<- fliplr(filter) -> (beg)", 
// i.e., the scalar product obtained by positioning 
// the flipped filter (formed into a row vector) 
// with the right side positioned on top of sample beg in data.

int filter_n_decimate(float *data,float *coef,int N,float *filter,int beg){
	int i,k;
//	float temp;
	k=0;
	for(i=beg;i<N+beg;i+=2){
/*		temp=0;
/*		for(j=i;j>i-2;j--)  // j=i, i-1
			temp+=data[j]*filter[i-j];*/
		coef[k] = data[i]*filter[0] + data[i-1]*filter[1];
//		coef[k]=temp;
		k++;
	}
	return(k);
}

// This routine is used in evaluating inverse wavelet transforms.
//
// Upsample and filter to implement a synthesis filter bank.
// Basically the inverse of filter_n_decimate() above.
// No actual upsampling etc., to avoid zero multiplies.
//
// coef is the input and contains coefficients resulting from 
// filtering and decimation by ofs. 
//
// Because this routine is typically called twice for inverse wavelet
// transforms, once for the low band and once for high, data is incremented via +=.
// Hence data should initially be set to 0.
//
// Viewed as a transform, filter_n_decimate above starts by putting the 
// "flipped" forward filter at begf, i.e.,  <- fliplr(ffilter) ->(begf)
// The resulting coefficient is at index 0.
// Here we are upsampling by ofs and filtering with the inv. filter
// time advanced by fbeg so the coefficient at index 0, multiplies the basis function
// (-fbeg)<-(ifilter)->(Nf-1-fbeg=beg).
// 
// How is that for confusing? All of this makes sense,
// when you write it down carefully. But who has the time ...:)

void upsample_n_filter(float *coef,float *data,int N,float *filter,int beg){
	int i,l,n,p,fbeg;
	float temp;

	fbeg=1-beg;

	for(i=0;i<N;i++){
		l=0;
		n=(i+fbeg)%2;
		p=(i+fbeg)/2;
		temp=0;
/*		for(j=n;j<2;j+=2) {
			temp+=coef[p-l]*filter[j];
			l++;
		}*/
		temp += coef[p-l] * filter[n];
		l++;
		// += for successive calls for low and high bands. 
		data[i]+=temp;
	}
}

// Used in forward wavelet trf.
//
// All rows of the image are run through the dual filterbank
// given by lp and hp. Results are returned in image "in place".
// low freq. coefficients start at 0 and end at N/2, where high freq.
// info starts.
void filt_n_dec_all_rows(float **image,int Ni,int Nj){
	int i,j,ofs;
	float *data, *ptr, *ptr0;
	data=allocate_1d_float((Nj+4),0);
	// Point in for required extensions.
	data+=2;

	// offset for the location where high band coefficients
	// should be copied.
	ofs=Nj>>1;

	for(i=0;i<Ni;i++){
		ptr = data; ptr0 = image[i];
		// Copy row.
		for(j=0;j<Nj;j++)
			*ptr++  = *ptr0++;
			//data[j]=image[i][j];

		extend_periodic(data,Nj);
		filter_n_decimate(data,image[i],Nj,MLP,begflp);
		filter_n_decimate(data,image[i]+ofs,Nj,MHP,begfhp);
	}

	data-=2;
	free((void *)data);
}

// Used in forward wavelet trf.
//
// Index swapped version of filt_n_dec_all_rows.
void filt_n_dec_all_cols(float **image,int Ni,int Nj){
	int i,j,ofs;
	float *data,*data2;

	data=allocate_1d_float((Ni+4),0);
	// Point in for required extensions.
	data+=2;

	// Need second array since cannot access columns directly via pointers.
	data2=allocate_1d_float(Ni,0);

	// offset for the location where high band coefficients
	// should be copied.
	ofs=Ni>>1;

	for(j=0;j<Nj;j++) {
		
		// Copy column.
		for(i=0;i<Ni;i++)
			data[i]=image[i][j];

		extend_periodic(data,Ni);

		filter_n_decimate(data,data2,Ni,MLP,begflp);
		filter_n_decimate(data,data2+ofs,Ni,MHP,begfhp);

		for(i=0;i<Ni;i++)
			image[i][j]=data2[i];
	}

	data-=2;
	free((void *)data);
	free((void *)data2);
}

// Used in inverse wavelet trf.
//
// All rows of the image are run through the dual synthesis filterbank
// given by lp and hp having number of taps Nl and Nh respectively. 
// Results are returned in image "in place".
//
// Here we actaully need to know the applied shift to the transform,
// hence the passed parameters lev and shift_arr.

void ups_n_filt_all_rows(float **image,int Ni,int Nj,int lev,int *shift_arr){
	int i,j,k,ofs;
	float *data1,*data2;

	ofs=Nj>>1;

	data1=allocate_1d_float((ofs+4),0);
	data2=allocate_1d_float((ofs+4),0);
	data1+=2;data2+=2;

	for(i=0;i<Ni;i++) {	

		for(j=0;j<ofs;j++) {

			k=j+ofs;
			// low pass and high pass coefficients.
			data1[j]=image[i][j];image[i][j]=0;
			data2[j]=image[i][k];image[i][k]=0;
		}

		// Take care of the extension at the boundaries.
		extend_periodic(data1,ofs);	
		extend_periodic(data2,ofs);	

		// invert low pass.
		upsample_n_filter(data1,image[i],Nj,MLP,begilp);
		// invert high pass.
		upsample_n_filter(data2,image[i],Nj,MHP,begihp);
	}

	data1-=2;data2-=2;
	free((void *)data1);
	free((void *)data2);
}

// Used in inverse wavelet trf.
//
// Index swapped version of ups_n_filt_all_rows.
void ups_n_filt_all_cols(float **image,int Ni,int Nj,int lev,int *shift_arr){
	int i,j,k,ofs;
	float *data1,*data2,*data3;

	ofs=Ni>>1;

	data1=allocate_1d_float((ofs+4),0);
	data2=allocate_1d_float((ofs+4),0);
	data1+=2;data2+=2;

	// Need third array since cannot access columns directly via pointers.
	data3=allocate_1d_float(Ni,0);

	for(j=0;j<Nj;j++) {	

		for(i=0;i<ofs;i++) {

			k=i+ofs;
			// low pass and high pass coefficients.
			data1[i]=image[i][j];
			data2[i]=image[k][j];
			data3[i]=data3[k]=0;
		}

		extend_periodic(data1,ofs);	
		extend_periodic(data2,ofs);	

		upsample_n_filter(data1,data3,Ni,MLP,begilp);
		upsample_n_filter(data2,data3,Ni,MHP,begihp);

		for(i=0;i<Ni;i++)
			image[i][j]=data3[i];
	}

	data1-=2;data2-=2;
	free((void *)data1);
	free((void *)data2);
	free((void *)data3);
}

